#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 17 12:27:44 2020

@author: ptarrio


  ; ***********************************************************************************************************************
  ; This procedure calculates the photometric redshift of a set of galaxies using
  ; the training set and the photometric redshift estimation method described in
  ; Tarrío and Zarattini 2020.
  ;
  ; *****************
  ; INPUT PARAMETERS:
  ; *****************
  ; - dir_name_cat_train: Full path and name of the training set (FITS file downloaded with this code)
  ; - dir_name_cat_test:  Full path and name of the set of galaxies for which we will calculate z_phot
  ;                       This set should be in FITS format. 
  ;                       It should contain a catalogue with any number of galaxies. 
  ;                       For each galaxy, it should provide the 5 features to be used for the computation 
  ;                       of the photometric redshift. Two options are admitted:
  ;                       - For aperture colours: R_KRON,GR_APER,RI_APER,IZ_APER,ZY_APER 
  ;                       - For Kron colours:     R_KRON,GR_KRON,RI_KRON,IZ_KRON,ZY_KRON   
  ;                       If a feature is not available, it should be set to -999
  ;                       The user may want to use prepare_test_galaxies.pro to create this file
  ; - dir_name_cat_out:   Full path and name of the output catalogue (where results will be saved)
  ;
  ; *****************
  ; KEYWORDS:
  ; *****************
  ; - feat_type:  Set of features to be used to compute the photometric redshift. Two options are available:
  ;               - 'aper': The 5 features to be used are the g-r, r-i, i-z, and z-y aperture colours and the r-band Kron magnitude 
  ;               - 'kron': The 5 features to be used are the g-r, r-i, i-z, and z-y Kron colours and the r-band Kron magnitude
  ;               The default value is 'aper'.
  ;               See Tarrío and Zarattini 2020 for details on how these colours are defined.
  ;
  ; - train_type: Defines the subset of the full training set that will be used to compute the photometric redshift when a
  ;               feature is missing from a galaxy. Two options are available:
  ;               - 'T5': Subset of the full training set that has the 5 features available
  ;               - 'T4': Subset of the full training set that has the same 4 features as the galaxy
  ;               The default value is 'T5'.
  ;               See Tarrío and Zarattini 2020 for details more details on these definitions.
  ;
  ; - std_type:   Defines how the standardization of the features is done. The standardization of the features
  ;               is done by subtracting the mean and dividing by the standard deviation of the features in a
  ;               given subset of the training set. This keyword defines the subset to be used for calculating
  ;               the mean and standard deviation to be applied in the standardization. Two options are available:
  ;               - 'T5': Subset of the full training set that has the 5 features available
  ;               - 'T1': For each feature, we take the subset of the full training set where that feature is available
  ;               The default value is 'T5'.
  ;               The standardization applies to training set galaxies as well as to the test galaxies.
  ;
  ; - use_5feat:  To compute (1) or not (0) the photometric redshift of galaxies that have the 5 features available.
  ;               The default value is 1
  ;
  ; - use_4feat:  To compute (1) or not (0) the photometric redshift of galaxies that have only 4 features available.
  ;               The default value is 1
  ;               
  ; - Nneigh:     Number of neighbours that will be used in the local linear regression (k in Tarrío and Zarattini 2020)
  ;               The default value is 100
  ;
  ; *****************
  ; OUTPUT:
  ; *****************
  ; The output catalogue contains the same number of galaxies as the input catalogue, in the same order.
  ; For each galaxy, the following fields are provided:
  ; - z_phot: photometric redshift
  ; - z_phot_err: error in the photometric redshift
  ; - n_feat: number of features that were used to calculate the photometric redshift
  ; - flag_extrapolate: flag to indicate if the photometric redshift is calculated via an interpolation (0)
  ;                     or an extrapolation (1) of the training features. Extrapolation occurs when the
  ;                     test galaxy lies outside the bounding box of its k nearest neighbours (in the feature space).
  ;                     See Tarrío and Zarattini 2020 for details more details on this flag.
  ;
  ; *****************
  ; RESTRICTIONS:
  ; *****************
  ; Not fully tested yet. Report any bugs to Paula Tarrío
  ; p.tarrio@oan.es
  ;
  ; *****************
  ; REVISION HISTORY
  ; *****************
  ; Written, Paula Tarrío, July 20 2020
  ;
  ; ***********************************************************************************************************************



"""

import numpy as np
from astropy.io import fits
from linear_regression_zphot_tarrio import linear_regression_zphot_tarrio


def compute_photo_z_pan_tarrio(dir_name_cat_train, dir_name_cat_test, dir_name_cat_out, feat_type='aper', train_type='T5', std_type='T5', use_5feat=1, use_4feat=1, Nneigh=100):

    # ----------------------------------------------------------
    # Checking keywords
    # ----------------------------------------------------------
    badinput = 0
    if (feat_type != 'aper' and feat_type != 'kron'):
        print('Error: The selected feat_type is not correct: It should be kron or aper (default).')
        badinput = 1

    if (train_type != 'T4' and train_type != 'T5'):
        print('Error: The selected train_type is not correct: It should be T4 or T5 (default).')
        badinput = 1

    if (std_type != 'T1' and std_type != 'T5'):
        print('Error: The selected std_type is not correct: It should be T1 or T5 (default).')
        badinput = 1

    if (use_5feat != 1 and use_5feat != 0):
        print('Error: The selected use_5feat is not correct: It should be 0 or 1 (default).')
        badinput = 1

    if (use_4feat != 1 and use_4feat != 0):
        print('Error: The selected use_4feat is not correct: It should be 0 or 1 (default).')
        badinput = 1

    if (Nneigh < 1):
        print('Error: The selected Nneigh is not correct: It should be at least 1 (100 recommended)')
        badinput = 1
    # ----------------------------------------------------------

    if badinput == 0:
        
        # Open training set
        cat = fits.getdata(dir_name_cat_train, 1) 
        Ntot = len(cat)
        
        # Select the features of the training set:
        if feat_type == 'aper':
            features_training_notnorm = np.concatenate((cat['gr_aper'][:, None],cat['ri_aper'][:, None],cat['iz_aper'][:, None],cat['zy_aper'][:, None],cat['r_kron'][:, None]),axis=1)
        if feat_type == 'kron':
            features_training_notnorm = np.concatenate((cat['gr_kron'][:, None],cat['ri_kron'][:, None],cat['iz_kron'][:, None],cat['zy_kron'][:, None],cat['r_kron'][:, None]),axis=1)
        
        # Subsets of the training set
        index_T5 = np.logical_and.reduce((features_training_notnorm[:,0] > -100 , features_training_notnorm[:,1] > -100 , features_training_notnorm[:,2] > -100 , features_training_notnorm[:,3] > -100 , features_training_notnorm[:,4] > -100),axis=0)
        index_T4_0123 = np.logical_and.reduce((features_training_notnorm[:,0] > -100 , features_training_notnorm[:,1] > -100 , features_training_notnorm[:,2] > -100 , features_training_notnorm[:,3] > -100 , features_training_notnorm[:,4] <= -100),axis=0)
        index_T4_0124 = np.logical_and.reduce((features_training_notnorm[:,0] > -100 , features_training_notnorm[:,1] > -100 , features_training_notnorm[:,2] > -100 , features_training_notnorm[:,3] <= -100 , features_training_notnorm[:,4] > -100),axis=0)
        index_T4_0134 = np.logical_and.reduce((features_training_notnorm[:,0] > -100 , features_training_notnorm[:,1] > -100 , features_training_notnorm[:,2] <= -100 , features_training_notnorm[:,3] > -100 , features_training_notnorm[:,4] > -100),axis=0)
        index_T4_0234 = np.logical_and.reduce((features_training_notnorm[:,0] > -100 , features_training_notnorm[:,1] <= -100 , features_training_notnorm[:,2] > -100 , features_training_notnorm[:,3] > -100 , features_training_notnorm[:,4] > -100),axis=0)
        index_T4_1234 = np.logical_and.reduce((features_training_notnorm[:,0] <= -100 , features_training_notnorm[:,1] > -100 , features_training_notnorm[:,2] > -100 , features_training_notnorm[:,3] > -100 , features_training_notnorm[:,4] > -100),axis=0)
        
        # Calculate the mean and standard deviation that will be used to standardize the features
        if std_type == 'T5':
            features_training_notnorm_formean = features_training_notnorm[index_T5]
            mu_features = features_training_notnorm_formean.mean(axis=0)
            sigma_features = features_training_notnorm_formean.std(axis=0)

        if std_type == 'T1':
            mu_gr = np.mean(features_training_notnorm[(features_training_notnorm[:,0] > -100), 0])
            mu_ri = np.mean(features_training_notnorm[(features_training_notnorm[:,1] > -100), 1])
            mu_iz = np.mean(features_training_notnorm[(features_training_notnorm[:,2] > -100), 2])
            mu_zy = np.mean(features_training_notnorm[(features_training_notnorm[:,3] > -100), 3])
            mu_r  = np.mean(features_training_notnorm[(features_training_notnorm[:,4] > -100), 4])
            sigma_gr = np.std(features_training_notnorm[(features_training_notnorm[:,0] > -100), 0])
            sigma_ri = np.std(features_training_notnorm[(features_training_notnorm[:,1] > -100), 1])
            sigma_iz = np.std(features_training_notnorm[(features_training_notnorm[:,2] > -100), 2])
            sigma_zy = np.std(features_training_notnorm[(features_training_notnorm[:,3] > -100), 3])
            sigma_r  = np.std(features_training_notnorm[(features_training_notnorm[:,4] > -100), 4])
            mu_features = [mu_gr,mu_ri,mu_iz,mu_zy,mu_r]
            sigma_features = [sigma_gr,sigma_ri,sigma_iz,sigma_zy,sigma_r]

        
        # We standardize the features of the training set:
        features_training = np.divide((features_training_notnorm - np.array([mu_features,]*Ntot)) , np.array([sigma_features,]*Ntot))

          
        # Open test set
        cat_test = fits.getdata(dir_name_cat_test, 1)
        Ntest = len(cat_test)
        
        if Ntest > 0:
        
            # Select features for the test galaxies
            if feat_type == 'aper':
                features_test_notnorm = np.concatenate((cat_test['gr_aper'][:, None],cat_test['ri_aper'][:, None],cat_test['iz_aper'][:, None],cat_test['zy_aper'][:, None],cat_test['r_kron'][:, None]),axis=1)
            if feat_type == 'kron':
                features_test_notnorm = np.concatenate((cat_test['gr_kron'][:, None],cat_test['ri_kron'][:, None],cat_test['iz_kron'][:, None],cat_test['zy_kron'][:, None],cat_test['r_kron'][:, None]),axis=1)
            # We normalize the features of the test set:
            features_test = np.divide((features_test_notnorm - np.array([mu_features,]*Ntest)) , np.array([sigma_features,]*Ntest))

            # Subsets of the test set
            good_5feat = np.logical_and.reduce((features_test[:,0] > -100 , features_test[:,1] > -100 , features_test[:,2] > -100 , features_test[:,3] > -100 , features_test[:,4] > -100),axis=0)
            good_0123 = np.logical_and.reduce((features_test[:,0] > -100 , features_test[:,1] > -100 , features_test[:,2] > -100 , features_test[:,3] > -100 , features_test[:,4] <= -100),axis=0)
            good_0124 = np.logical_and.reduce((features_test[:,0] > -100 , features_test[:,1] > -100 , features_test[:,2] > -100 , features_test[:,3] <= -100 , features_test[:,4] > -100),axis=0)
            good_0134 = np.logical_and.reduce((features_test[:,0] > -100 , features_test[:,1] > -100 , features_test[:,2] <= -100 , features_test[:,3] > -100 , features_test[:,4] > -100),axis=0)
            good_0234 = np.logical_and.reduce((features_test[:,0] > -100 , features_test[:,1] <= -100 , features_test[:,2] > -100 , features_test[:,3] > -100 , features_test[:,4] > -100),axis=0)
            good_1234 = np.logical_and.reduce((features_test[:,0] <= -100 , features_test[:,1] > -100 , features_test[:,2] > -100 , features_test[:,3] > -100 , features_test[:,4] > -100),axis=0)
            n_01234 = np.sum(good_5feat)
            n_0123 = np.sum(good_0123)
            n_0124 = np.sum(good_0124)
            n_0134 = np.sum(good_0134)
            n_0234 = np.sum(good_0234)
            n_1234 = np.sum(good_1234)

            # Linear regression to calculate z_phot for the galaxies in the test set that have the 5 features available
            if use_5feat == 1:
                index_training = index_T5
                if n_01234 > 0:
                    aux=1
                    result = linear_regression_zphot_tarrio(features_training[index_training,:], features_test[good_5feat,:], cat['z_spec'][index_training], Nneigh)
                    z_phot_5feat = result[0]
                    z_phot_err_5feat  = result[1]
                    flag_extrapolate_5feat = result[2]

        
            # Linear regression to calculate z_phot for the galaxies in the test set that have only 4 features available
            if use_4feat == 1:
                if train_type == 'T5':
                    index_training = index_T5
        
                ftr = features_training[:,[0,1,2,3]]
                fte = features_test[:,[0,1,2,3]]
                if train_type == 'T4':
                    index_training = index_T4_0123
                if n_0123 > 0:
                    result = linear_regression_zphot_tarrio(ftr[index_training,:], fte[good_0123,:], cat['z_spec'][index_training], Nneigh)
                    z_phot_0123 = result[0]
                    z_phot_err_0123  = result[1]
                    flag_extrapolate_0123 = result[2]
        
                ftr = features_training[:,[0,1,2,4]]
                fte = features_test[:,[0,1,2,4]]
                if train_type == 'T4':
                    index_training = index_T4_0124
                if n_0124 > 0:
                    result = linear_regression_zphot_tarrio(ftr[index_training,:], fte[good_0124,:], cat['z_spec'][index_training], Nneigh)
                    z_phot_0124 = result[0]
                    z_phot_err_0124  = result[1]
                    flag_extrapolate_0124 = result[2]
        
                ftr = features_training[:,[0,1,3,4]]
                fte = features_test[:,[0,1,3,4]]
                if train_type == 'T4':
                    index_training = index_T4_0134
                if n_0134 > 0:
                    result = linear_regression_zphot_tarrio(ftr[index_training,:], fte[good_0134,:], cat['z_spec'][index_training], Nneigh)
                    z_phot_0134 = result[0]
                    z_phot_err_0134  = result[1]
                    flag_extrapolate_0134 = result[2]
        
                ftr = features_training[:,[0,2,3,4]]
                fte = features_test[:,[0,2,3,4]]
                if train_type == 'T4':
                    index_training = index_T4_0234
                if n_0234 > 0:
                    result = linear_regression_zphot_tarrio(ftr[index_training,:], fte[good_0234,:], cat['z_spec'][index_training], Nneigh)
                    z_phot_0234 = result[0]
                    z_phot_err_0234  = result[1]
                    flag_extrapolate_0234 = result[2]
        
                ftr = features_training[:,[1,2,3,4]]
                fte = features_test[:,[1,2,3,4]]
                if train_type == 'T4':
                    index_training = index_T4_1234
                if n_1234 > 0:
                    result = linear_regression_zphot_tarrio(ftr[index_training,:], fte[good_1234,:], cat['z_spec'][index_training], Nneigh)
                    z_phot_1234 = result[0]
                    z_phot_err_1234  = result[1]
                    flag_extrapolate_1234 = result[2]

        
        
            # Define output catalogue structure
            col1 = fits.Column(name='z_phot', format='D', array=-np.ones(len(cat_test)))
            col2 = fits.Column(name='z_phot_err', format='D', array=-np.ones(len(cat_test)))
            col3 = fits.Column(name='n_feat', format='I', array=np.zeros(len(cat_test)))
            col4 = fits.Column(name='flag_extrapolate', format='I', array=-np.ones(len(cat_test)))
            hdu = fits.BinTableHDU.from_columns([col1 , col2 , col3 , col4])
            cat_out = hdu.data

            # Save results in output catalogue structure
            if use_5feat == 1:
                cat_out['z_phot'][good_5feat] = z_phot_5feat
                cat_out['z_phot_err'][good_5feat] = z_phot_err_5feat
                cat_out['n_feat'][good_5feat] = 5
                cat_out['flag_extrapolate'][good_5feat] = flag_extrapolate_5feat
            if use_4feat == 1:
                if n_0123 > 0:
                    cat_out['z_phot'][good_0123] = z_phot_0123
                    cat_out['z_phot_err'][good_0123] = z_phot_err_0123
                    cat_out['flag_extrapolate'][good_0123] = flag_extrapolate_0123
                    cat_out['n_feat'][good_0123] = 4
                if n_0124 > 0:
                    cat_out['z_phot'][good_0124] = z_phot_0124
                    cat_out['z_phot_err'][good_0124] = z_phot_err_0124
                    cat_out['flag_extrapolate'][good_0124] = flag_extrapolate_0124
                    cat_out['n_feat'][good_0124] = 4
                if n_0134 > 0:
                    cat_out['z_phot'][good_0134] = z_phot_0134
                    cat_out['z_phot_err'][good_0134] = z_phot_err_0134
                    cat_out['flag_extrapolate'][good_0134] = flag_extrapolate_0134
                    cat_out['n_feat'][good_0134] = 4
                if n_0234 > 0:
                    cat_out['z_phot'][good_0234] = z_phot_0234
                    cat_out['z_phot_err'][good_0234] = z_phot_err_0234
                    cat_out['flag_extrapolate'][good_0234] = flag_extrapolate_0234
                    cat_out['n_feat'][good_0234] = 4
                if n_1234 > 0:
                    cat_out['z_phot'][good_1234] = z_phot_1234
                    cat_out['z_phot_err'][good_1234] = z_phot_err_1234
                    cat_out['flag_extrapolate'][good_1234] = flag_extrapolate_1234
                    cat_out['n_feat'][good_1234] = 4
        
            # Save output catalogue in FITS format
            hdu.writeto(dir_name_cat_out, overwrite=True)
            

