#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 20 18:42:43 2020

@author: ptarrio

  ; ***********************************************************************************************************************
  ; This procedure puts the photometric information downloaded from Pan-STARRS for a given galaxy/galaxies
  ; in the appropriate format to be used as input of compute_photo_z_pan_tarrio.pro.
  ; For this, it takes as input the position (ra, dec) and magnitudes (r_kron, g, r, i, z, y) of the
  ; galaxies, corrects the magnitudes from REDDENING, computes the 5 features required by 
  ; compute_photo_z_pan_tarrio.pro, and saves the resulting information in an appropriate FITS file.
  ; 
  ; IMPORTANT: the magnitudes downloaded from Pan-STARRS should be taken from DR2 and STACK PHOTOMETRY.
  ;
  ; *****************
  ; INPUT PARAMETERS:
  ; *****************
  ; - ra: right ascension of the galaxy (in degrees). Vector with N elements (N=number of galaxies)
  ; - dec: declination of the galaxy (in degrees). Vector with N elements
  ; - r_kron: r-band Kron magnitude downloaded from Pan-STARRS (not dereddened). Vector with N elements
  ; - g: g-band (Kron or aperture) magnitude downloaded from Pan-STARRS (not dereddened). Vector with N elements
  ; - r: r-band (Kron or aperture) magnitude downloaded from Pan-STARRS (not dereddened). Vector with N elements
  ; - i: i-band (Kron or aperture) magnitude downloaded from Pan-STARRS (not dereddened). Vector with N elements
  ; - z: z-band (Kron or aperture) magnitude downloaded from Pan-STARRS (not dereddened). Vector with N elements
  ; - y: y-band (Kron or aperture) magnitude downloaded from Pan-STARRS (not dereddened). Vector with N elements
  ; - feat_type: Indicates the type of magnitudes that are being used. Two options are possible:
  ;               - 'aper': The g,r,i,z,y input magnitudes are aperture magnitudes
  ;               - 'kron': The g,r,i,z,y input magnitudes are Kron magnitudes
  ;               See Tarrío and Zarattini 2020 for details on how these magnitudes are defined.
  ; - dir_name_cat_test:  Full path and name of the test set (fits file that will be saved)
  ; - dir_name_ebvmap: Full name and path of the E(B-V) map from Schlegel et al 1998
  ;
  ; *****************
  ; OUTPUT:
  ; *****************
  ; The output catalogue contains the same number of galaxies as the input parameters, in the same order.
  ; It is saved in FITS format. For each galaxy, it provides the 5 features to be used for the computation
  ; of the photometric redshift. Two options are available:
  ; - For aperture colours: R_KRON,GR_APER,RI_APER,IZ_APER,ZY_APER
  ; - For Kron colours:     R_KRON,GR_KRON,RI_KRON,IZ_KRON,ZY_KRON
  ; If a feature is not available, it is set to -999
  ;
  ; *****************
  ; RESTRICTIONS:
  ; *****************
  ; Not fully tested yet. Report any bugs to Paula Tarrío
  ; p.tarrio@oan.es
  ;
  ; *****************
  ; REVISION HISTORY
  ; *****************
  ; Written, Paula Tarrío, July 20 2020
  ;
  ; ***********************************************************************************************************************


"""

import numpy as np
from astropy.io import fits
from astropy import units as u
from astropy.coordinates import SkyCoord
import healpy as hp
from dust_extinction.parameter_averages import F99

    
    
def prepare_test_galaxies_tarrio(ra, dec, r_kron, g, r, i, z, y, feat_type, dir_name_cat_test, dir_name_ebvmap):



    # Read the E(B-V) map for Schlegel et al 1998 
    # (downloaded from https://lambda.gsfc.nasa.gov/product/foreground/fg_sfd_get.cfm)
    ebvmap = hp.read_map(dir_name_ebvmap, nest=True)
    nside = hp.get_nside(ebvmap)

    # Calculate the Healpix pixels corresponding to the input coordinates
    coord = SkyCoord(ra=ra*u.degree, dec=dec*u.degree, frame='fk5')
    glon = coord.galactic.l.degree
    glat = coord.galactic.b.degree
    listpix = hp.ang2pix(nside, glon, glat, nest=True, lonlat=True)
    
    # Calculate E(B-V) values for each galaxy
    ebv = ebvmap[listpix]
    
    # We take A_lambda/E(B-V) for bands g,r,i,z from SDSS EDR paper (Stoughton et al. 2002), table 22
    Ag = ebv*3.793
    Ar = ebv*2.751
    Ai = ebv*2.086
    Az = ebv*1.479
    # We take A_lambda for band y from the effective lambda (Tonry et al 2012, table 4)
    # and the unreddening function of python that uses Fitzpatrick 1999 parametrization:
    R_v=3.1
    ext = F99(Rv=R_v) #Gives Alambda/Av   (Av = Rv * Ebv)
    Ay = ebv*ext(9620*u.AA)*R_v   # Ay = Ay/Av * Av = ext() * Rv * Ebv

    
    # Calculate dereddened magnitudes:
    r_kron_dered = r_kron - Ar
    g_dered = g - Ag
    r_dered = r - Ar
    i_dered = i - Ai
    z_dered = z - Az
    y_dered = y - Ay
      
     
    # Define fields of the output catalogue (the test set)
    col1 = fits.Column(name='r_kron', format='D', array=-999*np.ones(len(r_kron)))
    col2 = fits.Column(name='gr_kron', format='D', array=-999*np.ones(len(r_kron)))
    col3 = fits.Column(name='ri_kron', format='D', array=-999*np.ones(len(r_kron)))
    col4 = fits.Column(name='iz_kron', format='D', array=-999*np.ones(len(r_kron)))
    col5 = fits.Column(name='zy_kron', format='D', array=-999*np.ones(len(r_kron)))
    col6 = fits.Column(name='gr_aper', format='D', array=-999*np.ones(len(r_kron)))
    col7 = fits.Column(name='ri_aper', format='D', array=-999*np.ones(len(r_kron)))
    col8 = fits.Column(name='iz_aper', format='D', array=-999*np.ones(len(r_kron)))
    col9 = fits.Column(name='zy_aper', format='D', array=-999*np.ones(len(r_kron)))
    
    #  Create and fill the output catalogue (the test set)
    hdu = fits.BinTableHDU.from_columns([col1 , col2 , col3 , col4, col5 , col6 , col7 , col8, col9])
    cat_test = hdu.data
    cat_test['r_kron']  = r_kron_dered
    cat_test['r_kron'][(r_kron <= -100)] = -999
    
    if feat_type == 'kron':
        cat_test['gr_kron'] = g_dered-r_dered
        cat_test['ri_kron'] = r_dered-i_dered
        cat_test['iz_kron'] = i_dered-z_dered
        cat_test['zy_kron'] = z_dered-y_dered
        cat_test['gr_kron'][np.logical_or(g_dered <= -100 , r_dered <= -100)] = -999
        cat_test['ri_kron'][np.logical_or(r_dered <= -100 , i_dered <= -100)] = -999
        cat_test['iz_kron'][np.logical_or(i_dered <= -100 , z_dered <= -100)] = -999
        cat_test['zy_kron'][np.logical_or(z_dered <= -100 , y_dered <= -100)] = -999
    
    if feat_type == 'aper':
        cat_test['gr_aper'] = g_dered-r_dered
        cat_test['ri_aper'] = r_dered-i_dered
        cat_test['iz_aper'] = i_dered-z_dered
        cat_test['zy_aper'] = z_dered-y_dered
        cat_test['gr_aper'][np.logical_or(g_dered <= -100 , r_dered <= -100)] = -999
        cat_test['ri_aper'][np.logical_or(r_dered <= -100 , i_dered <= -100)] = -999
        cat_test['iz_aper'][np.logical_or(i_dered <= -100 , z_dered <= -100)] = -999
        cat_test['zy_aper'][np.logical_or(z_dered <= -100 , y_dered <= -100)] = -999
    
    # Save output catalogue in FITS format
    hdu.writeto(dir_name_cat_test, overwrite=True)


